import ProteinNetworks as PN
import pandas as pd
import stringdb
import networkx as nx
from umap import UMAP
import matplotlib.pyplot as plt
from math import log
from .wrappers import display_df, Check_Value, titler, Check_kwargs
from .mapping import get_mapping

INTERACTIONS_TYPE = {'STRINGdb':{
                        'physical': {'included_metrics': ['dscore', 'escore', 'tscore']}, 
                        'genetic': {'included_metrics': ['nscore', 'fscore', 'pscore', 'ascore']},
                        'all':{'included_metrics': ['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore']}},
                    }


def create_graph(geneList, interactions_type=None, **kwargs):
    """
    Function to create graph from gene list. It downloads interactions from STRING database and then create graph.
    For each gene, it creates node in graph. Then it creates edges between nodes if there are interactions between this genes.
    
    Parameters
    ----------
    geneList : pd.DataFrame
        pd.DataFrame contains ids in *protein_id_type* column
    interactions_type : str
        None by default. Which type of interactions you want to search. It`s ready to use presets. 'all', 'physical' or 'genetic'.
        Activated only if interactions_type is not None. When activated, included metrics is ignored    
    
    kwargs
    ----------
    **required_score** : *int.*
        Required score of interaction. From 150 to 999
    **taxId** : *int, optional.*
        Organism id. Human = 9606
    **included_metrics** : *list*, optional.
        List of metrics you want to include in total score. Default is ['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore']
    **base_degree** : *str or int or float, optional.*
        Base of degree in weights(x) function (see *get_edge_list()* function)
    **neg_exponent** : *str or int or float, optional.*
        Exponent of degree in weights(x) function (see *get_edge_list()* function)
        
    Returns
    ----------
    **NetworkAnalysis() object** contains graph with nodes from 'geneList'
    """

    # check correctness of kwargs
    valid_kwargs = {'taxId', 'required_score', 'included_metrics', 'base_degree', 'neg_exponent'}
    Check_kwargs(kwargs, valid_kwargs)
 
    required_score = kwargs.get('required_score', 400)
    taxId = kwargs.get('taxId', 9606)
    included_metrics = kwargs.get('included_metrics', ['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore'])
    base_degree = kwargs.get('base_degree', 'x')
    neg_exponent = kwargs.get('neg_exponent', 2)
    
    mapped_genes = get_mapping(geneList)
    network_obj = NetworkAnalysis(mapped_genes)
    network_obj.get_proteins_network(identifiers=network_obj.orig_data.stringId, 
                                     required_score=required_score, 
                                     included_metrics = included_metrics, 
                                     species=taxId,
                                     interactions_type=interactions_type)
    network_obj.get_edge_list(network_obj.proteins_network, id_type='GeneName', base_degree=base_degree, neg_exponent=neg_exponent)
    network_obj.create_graph_from_edge_list(network_obj.edge_list)
    return network_obj 

    

class NetworkAnalysis():
    
    id_type_converter = {'stringId': 'stringId', 'GeneName': 'preferredName'}
    def __init__(self, data, child_class=False):
        self.orig_data = data
        self.data = data.copy()
        if not child_class: # this class is parent for StringdbInteractions class (chech interaction.py). Bad solution, but it works    
            self.data.index = list(self.data.preferredName)
        self.edge_list = None
        self.graph = None
        self.id_type = None
        
    def _compute_prior_away(self, score, prior:float=0.041):

        """
        Compute a score with a prior probability subtracted.

        If the score is below the prior probability, it will be set to the prior probability.
        The score is then divided by (1 - prior) to scale.

        Parameters
        ----------
        score : float
            The score to be adjusted
        prior : float, optional
            The prior probability to subtract. Defaults to 0.041.

        Returns
        -------
        float
            The adjusted score
        """
        
        if score < prior: score = prior
        score_no_prior = (score - prior) / (1 - prior)

        return score_no_prior
    
    def calculate_total_score(self, data, included_metrics=['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore'], p:float=0.041):
        """
        Function for calculate total score
        Check list of metrix on: 
        https://string-db.org/help/faq/#the-protein-interactions-from-the-string-website-via-web-api-calls-what-do-the-score-columns-mean-for-example-nscore-fscore-tscore-etc

        Parameters
        ----------
        data : pd.DataFrame
            pd.DataFrame containing metrics
        included_metrics : list
            list of metrics you want to include in total score
        p : float
            prior probability, taken from STRING website. I think, better don`t touch this parameter

        Returns
        -------
        pd.Series
            pd.Series with total score
        """

        metrics_list = {'nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore'}
        for col in data.columns:
            Check_Value(col, metrics_list.union(['score']), ' ', message = f'Wrong column in data variable! \
                                            It must contain only {metrics_list.union(['score'])} columns')
        
        if type(included_metrics) not in [list]:
            raise Exception('Error of "included_metrics" variable. You must specify list of metrics you want to include in total score')
        for i in included_metrics:    
            Check_Value(i, metrics_list, ' ', message = f'Wrong element in "included_metrics" list! \
                                            It must contain only {metrics_list} metrics')

        cur_data = data[included_metrics]
        
        # calculate total score (https://string-db.org/help/faq/#how-are-the-scores-computed)
        s_tot_nop = cur_data.apply(lambda y: 1 - y.apply(lambda x: 1-self._compute_prior_away(x)).product(axis=0), axis=1)
        s_tot = s_tot_nop + p * (1 - s_tot_nop)

        return s_tot 
    
    def get_proteins_network(self, identifiers, 
                species=9606, 
                required_score=400,
                included_metrics=['dscore', 'escore', 'ascore', 'tscore'], 
                p:float=0.041,
                interactions_type=None):
    
        """
        Function for get proteins network

        Parameters
        ----------
        identifiers : list
            list of identifiers
        species : int
            species
        required_score : int
            required score
        included_metrics : list
            list of metrics you want to include in total score
        p : float
            prior probability, taken from STRING website. I think, better don`t touch this parameter

        Returns
        -------
        pd.DataFrame
            pd.DataFrame with network
        """
        if interactions_type is not None:
            Check_Value(interactions_type, {'genetic', 'physical', 'all'}, 'interactions_type')
            included_metrics = INTERACTIONS_TYPE['STRINGdb'][interactions_type]['included_metrics'] 
        
        proteins_network = stringdb.get_network(identifiers, species=species, required_score=required_score)
        # calculate total score taking into account "included_metrics"
        proteins_network['new_score'] = self.calculate_total_score(proteins_network.iloc[:, 6:], included_metrics, p)
        proteins_network.rename(mapper={'score': 'old_score', 'new_score': 'score'}, axis=1, inplace=True)

        proteins_network = proteins_network[proteins_network.score >= (required_score / 1000)]
        self.proteins_network = proteins_network
        #self.data['node'] = identifiers
        self.data['node'] = self.data['preferredName']

        return proteins_network
    
    
    def get_edge_list(self, network, base_degree='x', neg_exponent=2, id_type='stringId'):
        """
        Function for getting adjacency list from STRING network
        weights of edges are calculated as a function of 'score'= x: weight(x) = pow(base_degree, -neg_exponent)

        Weights of edges can be represented as an inverse power function 'x**(-a)', where 'a' is a real positive number \
                        or exponential function 'a**(-x)', where 'a' is a real positive number


        Example: get_edge_list(network, base_degree='x', neg_exponent=2) -> weights(x) = x**(-2)
                    get_edge_list(network, base_degree= np.e, neg_exponent=x) -> weights(x) = np.e**(x)
        
        
        Parameters
        ----------
        network : pd.DataFrame
            pd.DataFrame with STRING network
        base_degree : str or int or float
            base of degree in weights(x) function
        neg_exponent : str or int or float
            exponent of degree in weights(x) function
        id_type : str
            'stringId' or 'GeneName'
        
        Returns
        -------
        pd.DataFrame
            pd.DataFrame with adjacency list
        """

        # check parameters
        if not ((isinstance(base_degree, str) and base_degree == 'x' and isinstance(neg_exponent, (int, float)) and neg_exponent > 0) \
            or (isinstance(neg_exponent, str) and neg_exponent == 'x' and isinstance(base_degree, (int, float)) and base_degree > 0)):
            
            raise ValueError('Wrong parameters: "base_degree" must be "x" or real positive number, \
        "neg_exponent" must be real positive number or "x" respectively')

        Check_Value(id_type, {'stringId', 'GeneName'}, 'id_type')
        
        self.id_type = self.id_type_converter[id_type]
        if self.id_type == 'stringId': names = ['stringId_A', 'stringId_B']
        else: names = ['preferredName_A', 'preferredName_B']

        edge_list = pd.DataFrame()
        edge_list[['A', 'B']] = network[names]
        #prevent SettingWithCopyWarning message from appearing
        pd.options.mode.chained_assignment = None
        edge_list['weight'] = network['score'].apply(lambda x: pow(eval(str(base_degree)), eval(str('-') + str(neg_exponent))))
        self.edge_list = edge_list
        
        return edge_list

    def create_graph_from_edge_list(self, edge_list, weighted:bool=True):
        """
        Create graph from edge list

        Parameters
        ----------
        edge_list : pd.DataFrame
            pd.DataFrame with edges: 1st column - source vertex, 2nd column - target vertex, 3rd column - weight
        weighted : bool
            whether to use weighted edges

        Returns
        -------
        G : networkx.Graph
            networkx graph object
        """
        # Create an empty graph
        G = nx.Graph()

        # Add vertices to the graph
        vertices = set()
        for edge in edge_list.itertuples(index=False):
            vertices.add(edge[0])
            vertices.add(edge[1])
        G.add_nodes_from(vertices)

        # Add edges to the graph
        if weighted:
            for edge in edge_list.itertuples(index=False):
                G.add_edge(edge[0], edge[1], weight=edge[2])
        else: 
            for edge in edge_list.itertuples(index=False):
                G.add_edge(edge[0], edge[1], weight=1)
        
        self.graph = G
        
        return G
    
    def count_interactions(self):
        count_degree = pd.DataFrame.from_dict(dict(self.graph.degree), orient='index', columns=['degree']).reset_index(names=self.id_type)
        self.data = pd.merge(self.data, count_degree, how='outer', on=self.id_type)
        self.data.index = list(self.data.preferredName)
        count_degree = count_degree.rename(columns={'preferredName': 'protein', 'degree': '#interactions'})
        return count_degree.sort_values(by='#interactions', ascending=False)
    
    def clustering(self, method:str='louvain', weight:str='weight'):
        
        """
        Function for clustering proteins in graph using several methods.

        Parameters
        ----------
        method (str): Method for clustering. Available methods: 'louvain', 'infomap', 'edge_betweenness', 'fast_greedy'.
                      Default is 'louvain'.
        weight (str): Name of column in graph with weights. Default is 'weight'.
        
        Returns
        -------
        list of sets: List of clusters. Each cluster is set of protein IDs.
        """
        
        if method == 'louvain':
            clusters = nx.community.louvain_communities(self.graph, weight=weight)
        elif method == 'k_clique':
            clusters = nx.community.k_clique_communities(self.graph, k=4)
        elif method == 'greedy_modularity':
            clusters = nx.community.greedy_modularity_communities(self.graph, weight=weight)
        else: raise Exception('Wrong method name. Use: "louvain", k_clique", "greedy_modularity"')
        
        self.graph.clusters = clusters
        
        self.data['cluster'] = 0
        for i, c in enumerate(clusters):
            #self.data.loc[self.data[self.id_type].isin(clusters[i]), 'cluster'] = i + 1
            self.data.loc[self.data[self.id_type].isin(c), 'cluster'] = i + 1
            
            
        return clusters
    
    @staticmethod
    def _umap_layout(graph):
        dist_matrix = nx.to_numpy_array(graph)
        reducer = UMAP(n_components=2)
        umap_dist = reducer.fit_transform(dist_matrix) # position
        pos = {}
        for i, node in enumerate(graph.nodes()):
            pos[node] = umap_dist[i]
        return pos
    
    @staticmethod
    def _spectral_layout(graph):
        return nx.spring_layout(graph)
    
    @staticmethod
    def _spring_layout(graph):
        if nx.number_connected_components(graph) > 1:  
            return nx.spring_layout(graph, k=0.1)
        else:
            return nx.spring_layout(graph)
    
    def define_node_position(self, method='spring'):
        Check_Value(method, {'spring', 'umap', 'spectral'}, 'method')
        d_method = {'spring': self._spring_layout,
                'spectral': self._spectral_layout,
                'umap': self._umap_layout}
        try:
            self.graph.pos[method] = d_method[method](self.graph)
        except:
            self.graph.pos = {}
            self.graph.pos[method] = d_method[method](self.graph)
        
    def draw_graph(self, pos=None, clusterList=None, save=False, **kwargs):
        """
        Function for fast graph visualization

        Parameters
        ----------
        graph: networkx.Graph
            Graph object
        clusterList: pd.Series object
            List of clusters from NetworkAnalysis.data.cluster object. Contains indexes of nodes in the graph
        **kwargs: dict
            valid kwargs:
    
            *figsize*: tuple (width, height). Size of the figure
    
            *node_size*: int. Size of the nodes
    
            *node_color*: string. Color of node. Default: '#B897D5'
    
            *font_size*: int. Size of the font. Default: 5
    
            *font_color*: string. Color of the font. Default: '#222222'
    
            *edge_color*: string. Color of the edge. Default: 'lightgray'
            
            *palette*: string. Choose one of this color palettes for coloring clusters: https://matplotlib.org/stable/users/explain/colors/colormaps.html
            
            *dpi*: int. Work with 'save'=True. Dpi of saving image
            
        Returns
        -----------
        None
        
        """
        valid_kwargs = {'figsize','node_size', 'node_color', 'node_edge_color', 'edge_color', 'font_size', 'font_color', 'view_labels', 'palette', 'pos', 'dpi'}
        Check_kwargs(kwargs, valid_kwargs)
    
        figsize = kwargs.get('figsize', (10, 10))
        node_size = kwargs.get('node_size', 50)
        unicolor_node_color = kwargs.get('node_color', '#B897D5')
        font_size = kwargs.get('font_size', 5)
        font_color = kwargs.get('font_color', '#222222')
        edge_color = kwargs.get('edge_color', 'lightgray')
        view_labels = kwargs.get('view_labels', True)
        palette = kwargs.get('palette', 'rainbow')
        dpi = kwargs.get('dpi', 400)
    
        
        if clusterList is not None:
            clusters = list(map(lambda x: clusterList[clusterList.index == x].iloc[0], self.graph.nodes()))
            cmap = plt.get_cmap(palette)  # Choose a color palette
            num_clusters = len(clusterList.unique())
            node_colors = [cmap(i / num_clusters) for i in clusters]
        else:
            clusters = [1]*len(self.graph.nodes())
            node_colors = [unicolor_node_color]*len(self.graph.nodes())

        fig = plt.figure(figsize=figsize)
        
        if pos is None:
            pos = self._spring_layout(self.graph)
    
        nx.draw_networkx_nodes(self.graph, 
                                pos, 
                                node_size=node_size, 
                                node_color=node_colors)

        nx.draw_networkx_edges(self.graph, pos, alpha=0.5, edge_color=edge_color)
        
        if view_labels:
            for node in self.graph.nodes:
                plt.text(pos[node][0], pos[node][1], 
                        node, 
                        ha='center', va='center',
                        fontsize=font_size,
                        color=font_color,
                        fontweight='bold')

        plt.axis('off')
        if save:
            plt.savefig('graph.png', dpi=dpi)
        plt.show()
        
    def visualize_graph(self, pos_method='spring', view_clusters=False, clustering_method='louvain', **kwargs):
        
        """
        Function for graph visualization
        
        Parameters
        ----------
        pos_method: str
            Method for node positioning. Choose one of: 'spring', 'umap', 'spectral'
        view_clusters: bool
            If True, first, clustering will be performed with 'clustering_method' method and then cluster colors will be used for visualization
        clustering_method: str
            Choose one of the following clustering methods: 'louvain', 'k_clique', 'greedy_modularity'
        **kwargs: dict
            valid kwargs for 'draw_graph' method
        
        Returns
        -------
        None
        """
        try:
            pos = self.graph.pos[pos_method]
        except:
            self.define_node_position(method=pos_method)
            pos = self.graph.pos[pos_method]
        if view_clusters:
            self.clustering(method=clustering_method)
            self.draw_graph(pos=pos, clusterList=self.data.cluster, **kwargs)
        else:
            self.draw_graph(pos=pos, **kwargs)


#from .STRING_enrichment import *
#from .R_requests import *
from . import R_requests, enrichment, interactions, mapping, networks, wrappers
from .wrappers import save_table, display_df
from .enrichment import get_enrichment, EnrichmentAnalysis
from .mapping import get_mapping, STRING_mapping
from .networks import NetworkAnalysis, create_graph
from .interactions import get_interactors_from_biogrid, get_interactors_from_stringdb, \
    get_interactionsTable_from_biogrid, get_interactionsTable_from_stringdb, \
        merging_interactors_stringdb_and_biogrid, get_interactors
        
    



from   .networks import NetworkAnalysis
from   .wrappers import Check_Value, Check_type, Check_kwargs

from biogridpy.biogrid_client import BioGRID
import os
import pandas as pd
import stringdb
import numpy as np
from io import StringIO


BIOGRID_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'BioGRID')

conf_path = os.path.join(BIOGRID_PATH, 'biogridpyrc')
INTERACTIONS_TYPE = {'STRINGdb':{
                        'physical': {'included_metrics': ['dscore', 'escore', 'tscore']}, 
                        'genetic': {'included_metrics': ['nscore', 'fscore', 'pscore', 'ascore']},
                        'all':{'included_metrics': ['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore']}},
                    'BioGRID':{
                        'physical': {'EXPERIMENTAL_SYSTEM_TYPE': 'physical'}, 
                        'genetic': {'EXPERIMENTAL_SYSTEM_TYPE': 'genetic'},
                        'all': {'EXPERIMENTAL_SYSTEM_TYPE': 'any'}} 
                    }

def get_interactionsTable_from_biogrid(geneList:list,
                                       taxId=9606,
                                       interactions_type = None,
                                       EXPERIMENTAL_SYSTEM_TYPE = 'any',
                                       required_evidence = 1):
    """
    Function returns interaction table from BioGRID for given list of genes and organism id (taxId).

    Parameters
    ----------
    geneList : list
        list of genes
    taxId : int
        organism id. Human = 9606
    EXPERIMENTAL_SYSTEM_TYPE : str
            type of experimental system. Valid types are 'genetic', 'physical' or 'any'.
            Default is 'any'
    required_evidence : int
        required number of evidence for interaction
        Default is 1. It means that BioGRID has 1 mention about interaction between two proteins
    interactions_type : str
        None by default. Which type of interactions you want to search. It`s ready to use presets. 'all', 'physical' or 'genetic'.
        Activated only if interactions_type is not None. When activated, EXPERIMENTAL_SYSTEM_TYPE is ignored

    Returns
    -------
    pd.DataFrame
        interaction table
    """
    if interactions_type is not None:
        Check_Value(interactions_type, {'genetic', 'physical', 'all'}, 'interactions_type')
        EXPERIMENTAL_SYSTEM_TYPE = INTERACTIONS_TYPE['BioGRID'][interactions_type]['EXPERIMENTAL_SYSTEM_TYPE']
    BGI = BiogridInteractions(geneList, taxId)
    BGI.get_interaction_partners(EXPERIMENTAL_SYSTEM_TYPE, required_evidence = required_evidence)
    return BGI.interactionTable

def get_interactors_from_biogrid(geneList:list, 
                                 taxId=9606, 
                                 EXPERIMENTAL_SYSTEM_TYPE = 'any',
                                 required_evidence = 1,
                                 interactions_type = None)->list:
    
    """
    Function returns list of interactors from BioGRID for given list of genes and organism id (taxId).

    Parameters
    ----------
    geneList : list
        list of genes
    taxId : int
        organism id. Human = 9606
    EXPERIMENTAL_SYSTEM_TYPE : str
        type of experimental system. Valid types are 'genetic', 'physical' or 'any'.
        Default is 'any'
    required_evidence : int
        required number of evidence for interaction
        Default is 1. It means that BioGRID has 1 mention about interaction between two proteins
    interactions_type : str
        None by default. Which type of interactions you want to search. It`s ready to use presets. 'all', 'physical' or 'genetic'.
        Activated only if interactions_type is not None. When activated, EXPERIMENTAL_SYSTEM_TYPE isignored

    Returns
    -------
    pd.DataFrame
        1-column DataFrame of interactors (header = Gene)
    """
    Check_Value(EXPERIMENTAL_SYSTEM_TYPE, {'any', 'genetic', 'physical'}, 'EXPERIMENTAL_SYSTEM_TYPE')
    if interactions_type is not None:
        Check_Value(interactions_type, {'genetic', 'physical', 'all'}, 'interactions_type')
        EXPERIMENTAL_SYSTEM_TYPE = INTERACTIONS_TYPE['BioGRID'][interactions_type]['EXPERIMENTAL_SYSTEM_TYPE']
        
    BGI = BiogridInteractions(geneList, taxId)
    BGI.get_interaction_partners(EXPERIMENTAL_SYSTEM_TYPE, required_evidence = required_evidence)
    return BGI.interactorsList


def get_interactionsTable_from_stringdb(geneList:list, 
                                        required_score=400,
                                        included_metrics=['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore'],
                                        limit=None,
                                        taxId=9606,
                                        interactions_type = None)->pd.DataFrame:
  
    """
    Function returns interaction table from STRINGdb for given list of genes and organism id (taxId) and
    required score of interaction.

    Parameters
    ----------
    geneList : list
        list of genes
    required_score : int
        required score of interaction. From 150 to 999
    limit : int or None
        limit of interactors for each gene from geneList
    taxId : int
        organism id. Human = 9606
    interactions_type : str
        None by default. Which type of interactions you want to search. It`s ready to use presets. 'all', 'physical' or 'genetic'.
        Activated only if interactions_type is not None. When activated, included metrics is ignored

    Returns
    -------
    pd.DataFrame
        interaction table
    """
    if interactions_type is not None:
        Check_Value(interactions_type, {'genetic', 'physical', 'all'}, 'interactions_type')
        included_metrics = INTERACTIONS_TYPE['STRINGdb'][interactions_type]['included_metrics']
    STRINGi = StringdbInteractions(geneList, taxId)
    STRINGi.get_interaction_partners(required_score=required_score, limit=limit, included_metrics=included_metrics)
    return STRINGi.interactionTable

def get_interactors_from_stringdb(geneList:list,
                                  required_score=400,
                                  included_metrics=['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore'],
                                  limit=None,
                                  taxId=9606,
                                  interactions_type = None)->pd.DataFrame:
    """
    Function returns list of interactors from STRINGdb for given list of genes.

    Parameters
    ----------
    geneList : list
        list of genes
    required_score : int
        required score of interaction. From 150 to 999
    limit : int or None
        limit of interactors for each gene from geneList
    taxId : int
        organism id. Human = 9606
    interactions_type : str
        None by default. Which type of interactions you want to search. It`s ready to use presets. 'all', 'physical' or 'genetic'.
        Activated only if interactions_type is not None. When activated, included metrics is ignored

    Returns
    -------
    pd.DataFrame
        1-column DataFrame of interactors (header = Gene)
    """

    if interactions_type is not None:
        Check_Value(interactions_type, {'genetic', 'physical', 'all'}, 'interactions_type')
        included_metrics = INTERACTIONS_TYPE['STRINGdb'][interactions_type]['included_metrics']    
    
    STRINGi = StringdbInteractions(geneList, taxId)
    STRINGi.get_interaction_partners(required_score=required_score, limit=limit, included_metrics=included_metrics)
    return STRINGi.interactorsList

def merging_interactors_stringdb_and_biogrid(geneList1, geneList2, method= 'union', mergeCol='Gene'):
    
    """
    Function merges two lists of interactors (from STRINGdb and BioGRID) using union or intersection operation.

    Parameters
    ----------
    geneList1 : pd.DataFrame, list, tuple, pd.Series
        list of genes from STRINGdb
    geneList2 : pd.DataFrame, list, tuple, pd.Series
        list of genes from BioGRID
    method : str
        method of merging. 'union' or 'intersection'. 'union' by default
    mergeCol : str
        column name to merge. 'Gene' by default

    Returns
    -------
    pd.DataFrame
        merged list of interactors
    """
    Check_Value(method, {'union', 'intersection'}, 'method')
    Check_type(geneList1, (pd.DataFrame, pd.Series, list, tuple), 'geneList1')
    Check_type(geneList2, (pd.DataFrame, pd.Series, list, tuple), 'geneList2')
    
    if isinstance(geneList1, (list, pd.Series, tuple)):
        geneList1 = pd.DataFrame({'Gene': geneList1})  
        
    if isinstance(geneList2, (list, pd.Series, tuple)):
        geneList2 = pd.DataFrame({'Gene': geneList2})
        
    method_key = {'union': 'outer', 'intersection': 'inner'}
    return pd.merge(geneList1, geneList2, how=method_key[method], on=mergeCol)

def get_interactors(target_genes, 
                    method= 'union',
                    interactions_type = None, 
                    **kwargs):
    
    """
    Function returns list of interactors for given list of genes.
    
    Parameters
    ----------
    target_genes : list
        list of genes
    method : str
        method of merging. 'union' or 'intersection'. 'union' by default
    interactions_type : str
        None by default. Which type of interactions you want to search. It`s ready to use presets. 'all', 'physical' or 'genetic'.
        Activated only if interactions_type is not None. When activated, included metrics and EXPERIMENTAL_SYSTEM_TYPE are ignored
    **kwargs :
        parameters for fine-tuning the interactome search
    
    kwargs
    ----------
    **included_metrics** : *list*, optional.
        list of metrics you want to include in total score. Default is ['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore']
    **taxId** : *int, optional*.
        organism id. Human = 9606. Default is 9606
    **EXPERIMENTAL_SYSTEM_TYPE** : *str, optional*.
        type of experimental system. 'physical' or 'genetic'. 'physical' by default
    **required_evidence** : *int, optional*.
        required number of evidence for interaction
        Default is 1. It means that BioGRID has 1 mention about interaction between two proteins
              
    **required_score** : *int, optional*.
        required score of interaction. From 150 to 999
    **limit** : *int or None, optional*.
        limit of interactors for each gene from geneList
    **taxId** : *int, optional*.
        organism id. Human = 9606   
         
    **mergeCol** : *str*.
        column name to merge. 'Gene' by default
    
    Returns
    -------
    pd.DataFrame
        1-column DataFrame of interactors (header = 'Gene' or choosen value of 'mergeCol')
    """
    # check correctness of kwargs
    valid_kwargs = {'EXPERIMENTAL_SYSTEM_TYPE', 'taxId', 'required_score', 'required_evidence', 'included_metrics', 'limit', 'mergeCol'}
    Check_kwargs(kwargs, valid_kwargs)
 
    EXPERIMENTAL_SYSTEM_TYPE = kwargs.get('EXPERIMENTAL_SYSTEM_TYPE', 'any')
    required_score = kwargs.get('required_score', 400)
    taxId = kwargs.get('taxId', 9606)
    required_evidence = kwargs.get('required_evidence', 1)
    limit = kwargs.get('limit', None)
    included_metrics = kwargs.get('included_metrics', ['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore'])
    mergeCol = kwargs.get('mergeCol', 'Gene')
    
    #find interactors from biogrid and stringdb
    geneList_biogrid = get_interactors_from_biogrid(target_genes, 
                                                    EXPERIMENTAL_SYSTEM_TYPE=EXPERIMENTAL_SYSTEM_TYPE,
                                                    interactions_type=interactions_type, 
                                                    taxId=taxId,
                                                    required_evidence=required_evidence)
    geneList_string = get_interactors_from_stringdb(target_genes, 
                                                    required_score=required_score, 
                                                    included_metrics=included_metrics, 
                                                    taxId=taxId,
                                                    limit=limit,
                                                    interactions_type=interactions_type)
    # merge lists of interactors
    interactorsList = merging_interactors_stringdb_and_biogrid(geneList_biogrid, geneList_string, method=method, mergeCol=mergeCol)
    print(f'The list of interactors was successfully created.\nNumber of participants: {len(interactorsList)}')
    return interactorsList

class BiogridInteractions():
    
    def __init__(self, geneList:list, taxId=9606):
        
        """
        BiogridInteractions class conctructor.
        
        Parameters
        ----------
        geneList : list
            list of genes
        taxId : int
            organism id. Human = 9606
        """
        if not isinstance(geneList, list):
            raise Exception('Wrong type of "geneList". It must be "list"')
        self.geneList = geneList
        self.taxId = taxId
        self.interactionTable = None
        self.interactorsList = None
    
    @staticmethod    
    def _calculate_evidence(table):
        table['Evidence'] = 0
        for id in table.OFFICIAL_SYMBOL_A.unique():
            subframe = table[table.OFFICIAL_SYMBOL_A == id]
            evidence_col = subframe.apply(lambda x: subframe.OFFICIAL_SYMBOL_B[subframe.OFFICIAL_SYMBOL_B == x.OFFICIAL_SYMBOL_B].count(), axis=1)
            table.update({'Evidence': evidence_col})
        return table['Evidence']    
    
    def get_interaction_partners(self, EXPERIMENTAL_SYSTEM_TYPE = 'any', required_evidence = 1):
        
        """
        Function for getting interaction data from BioGRID database
        
        Parameters
        ----------
        EXPERIMENTAL_SYSTEM_TYPE : str, optional
            type of experimental system. Valid types are 'genetic', 'physical' or 'any'.
            Default is 'any'
        
        Returns
        -------
        pd.DataFrame
            DataFrame containing all possible interactions between genes in geneList
        """
        
        Check_type(required_evidence, (int, float), 'required_evidence')
        Check_Value(EXPERIMENTAL_SYSTEM_TYPE, {'any', 'genetic', 'physical'}, 'EXPERIMENTAL_SYSTEM_TYPE')
        if EXPERIMENTAL_SYSTEM_TYPE == 'any':
            EXPERIMENTAL_SYSTEM_TYPE = ['genetic', 'physical']
        elif EXPERIMENTAL_SYSTEM_TYPE == 'genetic':
            EXPERIMENTAL_SYSTEM_TYPE = ['genetic']
        elif EXPERIMENTAL_SYSTEM_TYPE == 'physical':
            EXPERIMENTAL_SYSTEM_TYPE = ['physical']

        BG = BioGRID(config_filepath=conf_path)
        bg_interact = BG.interactions('json', geneList=self.geneList,
                                        includeEvidence='true',
                                        taxId=self.taxId)
        
        interactionTable = pd.read_json(StringIO(bg_interact.toDataFrame()), orient='index')
        
        interactionTable = self._target_gene_1st_col(interactionTable)
        interactionTable = interactionTable[interactionTable.OFFICIAL_SYMBOL_A.isin(self.geneList) | interactionTable.OFFICIAL_SYMBOL_B.isin(self.geneList)]
        interactionTable['Evidence'] = self._calculate_evidence(interactionTable) # calculation of evidence
        interactionTable.sort_values(by='QUANTITATION', ascending=False, inplace=True) # sorting by QUANTITATION
        interactionTable = interactionTable[(interactionTable.ORGANISM_A == self.taxId) & (interactionTable.ORGANISM_B == self.taxId)] # keep only right taxId
        interactionTable.drop_duplicates(subset=['OFFICIAL_SYMBOL_A', 'OFFICIAL_SYMBOL_B'], inplace=True) # drop duplicates
        
        # filter by EXPERIMENTAL_SYSTEM_TYPE and evidence
        interactionTable = interactionTable[interactionTable.EXPERIMENTAL_SYSTEM_TYPE.isin(EXPERIMENTAL_SYSTEM_TYPE)]
        interactionTable = interactionTable[interactionTable.Evidence >= required_evidence]
        
        self.interactionTable = interactionTable
        self.interactorsList = pd.DataFrame({'Gene': np.unique(np.concatenate((interactionTable.OFFICIAL_SYMBOL_A.unique(), \
            interactionTable.OFFICIAL_SYMBOL_B.unique())))})
                        
    def _swap_cols(self,row):
        
        """
        Function for swapping columns in a row of interaction table
        
        Parameters
        ----------
        row : pd.Series
            row of interaction table
            
        Returns
        -------
        pd.Series
            row of interaction table with swapped columns
        """
        
        A_cols = ['ENTREZ_GENE_A', 'BIOGRID_ID_A', 'SYSTEMATIC_NAME_A', 'OFFICIAL_SYMBOL_A', 'SYNONYMS_A', 'ORGANISM_A']
        B_cols = ['ENTREZ_GENE_B', 'BIOGRID_ID_B', 'SYSTEMATIC_NAME_B', 'OFFICIAL_SYMBOL_B', 'SYNONYMS_B', 'ORGANISM_B']
        
        k = {}
        for a, b in zip(A_cols, B_cols):
            k[b] = a
            k[a] = b
        new_row = row.rename(k)
        return new_row
    
    def _target_gene_1st_col(self, table):
        
        """
        The function moves target genes to column "*_A" and their interactors to column "*_B" in interaction table
        
        Parameters
        ----------
        table : pd.DataFrame
            interaction table
            
        Returns
        -------
        pd.DataFrame
            table with target genes in column A
        """

        all_cols = table.columns
        table = table.apply(lambda x: self._swap_cols(x) if x.OFFICIAL_SYMBOL_A not in self.geneList else x, axis=1)
        return table.reindex(columns=all_cols)


class StringdbInteractions():
    
    def __init__(self, geneList:list, taxId:int=9606):
                
        """
        StringdbInteractions class conctructor.
        
        Parameters
        ----------
        geneList : list
            list of genes
        taxId : int
            organism id. Human = 9606
        """
        if not isinstance(geneList, list):
            raise Exception('Wrong type of "geneList". It must be "list"')
        
        self.geneList = geneList
        self.taxId = taxId
        self.interactionTable = None
        self.interactorsList = None
        self.parent = NetworkAnalysis(data=pd.DataFrame(), child_class=True)
    
    def _compute_prior_away(self, score, prior:float=0.041):
        """
        Compute a score with a prior probability subtracted.

        If the score is below the prior probability, it will be set to the prior probability.
        The score is then divided by (1 - prior) to scale.

        Parameters
        ----------
        score : float
            The score to be adjusted
        prior : float, optional
            The prior probability to subtract. Defaults to 0.041.

        Returns
        -------
        float
            The adjusted score
        """
        return self.parent._compute_prior_away(score=score, prior=prior)
        
    
    def calculate_total_score(self, data, included_metrics=['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore'], p:float=0.041):
        """
        Function for calculate total score
        Check list of metrix on: 
        https://string-db.org/help/faq/#the-protein-interactions-from-the-string-website-via-web-api-calls-what-do-the-score-columns-mean-for-example-nscore-fscore-tscore-etc

        Parameters
        ----------
        data : pd.DataFrame
            pd.DataFrame containing metrics
        included_metrics : list
            list of metrics you want to include in total score
        p : float
            prior probability, taken from STRING website. I think, better don`t touch this parameter

        Returns
        -------
        pd.Series
            pd.Series with total score
        """
        return self.parent.calculate_total_score(data=data, included_metrics=included_metrics, p=p) 
        
    def get_interaction_partners(self, 
                                 required_score:int=400,
                                 limit:int=None,
                                 included_metrics=['nscore', 'fscore', 'pscore', 'dscore', 'escore', 'ascore', 'tscore'],
                                 p:float=0.041):
        
        
        """
        Function returns interaction table from stringdb for given list of genes and organism id (taxId) and
        required score of interaction.
        
        Parameters
        ----------
        required_score : int
            required score of interaction. From 150 to 999
        limit : int or None
            limit of interactors for each gene from geneList
        included_metrics : list
            list of metrics you want to include in total score
            For example, physical metrix: ['dscore', 'escore', 'ascore', 'tscore']
        p : float
            prior probability, taken from STRING website. I think, better don`t touch this parameter
        
        Returns
        -------
        None
        """
        # for some reason, if you search partners only for 1 gene, limit=10 by default, but there are no limits for several genes
        if len(self.geneList) == 1 and limit is None:
            limit = 1000000
        
        interactionTable = stringdb.get_interaction_partners(self.geneList, required_score=required_score, limit=limit)
                
        #cutoff genes with new_score less than required_score. new_score formed by included_metrix
        interactionTable['new_score'] = self.calculate_total_score(data=interactionTable.iloc[:, 6:], included_metrics=included_metrics, p=p)
        interactionTable.rename(mapper={'score': 'old_score', 'new_score': 'score'}, axis=1, inplace=True)
        interactionTable = interactionTable[interactionTable.score >= (required_score / 1000)]
        
        self.interactionTable=interactionTable
        self.interactorsList = pd.DataFrame({'Gene': np.unique(np.concatenate((self.interactionTable.preferredName_A.unique(), \
            self.interactionTable.preferredName_B.unique())))})  # create DataFrame of unique Genes. It`s strange, but faster
    
    
    





from   .R_requests import Check_R_packages, short_R_output
from   .wrappers import Check_Value, display_df, titler, print_upline, print_downline, save_table, create_subframe_by_names

from   datetime import datetime
import functools
from   math import log
import os
import pandas as pd
import stringdb
from   subprocess import Popen, PIPE
from   tabulate import tabulate


# path to directory contains all RScripts
RSCRIPTS_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'RScripts')


def get_enrichment(proteins, protein_id_type='Gene', species=9606, silent=True):
    """
    Function for one-click presetted enrichment analysis. It creates EnrichmentAnalysis object with your data,
    drops duplicates, maps protein IDs to STRING IDs, makes enrichment analysis and shows
    enrichment categories.

    Parameters
    ----------
    proteins : pd.DataFrame
        DataFrame with protein IDs. It must contain either a "Gene" or "UniProtID" column
    protein_id_type : str
        Type of protein ID. Valid types: 'Gene', 'UniProtID'
    species : int
        Species ID. For example, human species ID = 9606
    silent : bool
        If True, then function will not print anything

    Returns
    -------
    EnrichmentAnalysis
        EnrichmentAnalysis object
    """
    obj = EnrichmentAnalysis(proteins, protein_id_type=protein_id_type)
    obj.drop_duplicated_genes(silent=silent) # drop duplicates from your protein id set
    obj.get_mapped(species=species) # find STRINGid for each protein id
    obj.get_enrichment() # make enrichment
    obj.show_enrichment_categories() # get enrichment category
    return obj


class EnrichmentAnalysis:
    types = {'UniProtID': 'queryItem', 'Gene': 'preferredName'}

    def __init__(self, data, enrichment = None, protein_id_type='Gene'):
        """
        EnrichmentAnalysis class conctructor.

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe containing the protein ID for analysis. It must contain either a "Gene" or "UniProtID" column
        enrichment : pd.DataFrame, optional
            Dataframe containing the results of previous enrichment analysis
        protein_id_type : str, optional
            type of protein ID. Valid Types

        Returns
        -------
        None
        """
        #check correctness of inputs
        self.protein_id_type = protein_id_type
        self._check_proteins_column(data)

        self.orig_data = data
        self.proteins = self.orig_data[self.protein_id_type]
        self.enrichment = enrichment # enrichment from previous analysis

    def _check_proteins_column(self, data):
        """
        The function checks the presence of the 'Gene' and 'UniProtID' columns in data

        Parameters
        ----------
        data : pd.DataFrame
            data contains proteins` IDs

        Returns
        -------
        None
        """

        valid_cols = set(self.types.keys()).intersection(data.columns)
        if len(valid_cols) == 0:
            message = 'The protein data must contain either a "Gene" or "UniProtID" column'
            del self
            raise Exception(message)
        elif len(valid_cols) == 1:
            p_id = next(iter(valid_cols))
            if self.protein_id_type != p_id:
                self.protein_id_type = p_id
                print(f'You choose "protein_id_type" that wasn`t contained in your data. '
                      f'"protein_id_type" is changed to "{self.protein_id_type}"\n')
        elif len(valid_cols) == 2:
            pass

    def _get_valid_category(self)->set:
        """
        function return set of valid category names for current enrichment analysis
        :return:
        """
        return set(self.enrichment.category.unique())

    def _find_nomapped_genes(self):
        """
        check genes in dataset which didn't find by STRING (nomapped genes) and
        found by STRING but wasn`t in dataset (overmapped genes)
        
        Parameters
        ----------
        None

        Returns
        -------
        tuple
            tuple of two sorted lists. First list contains nomapped genes and second list contains overmapped genes.
        """
        nomapped = sorted(
            set(self.proteins.unique()).difference(set(self.genes_mapped[self.types[self.protein_id_type]].unique())))
        overmapped = sorted(
            set(self.genes_mapped[self.types[self.protein_id_type]].unique()).difference(set(self.proteins.unique())))
        return nomapped, overmapped
    
    @titler('DISCARDING DUPLICATES')
    def drop_duplicated_genes(self, silent=False):
        """
        Function for droppig dublicated genes
        
        Parameters
        ----------
        subset : list, optional
            Only consider certain columns for identifying duplicates, by default use all columns.

        Returns
        -------
        pd.DataFrame
            df of dropped genes
        """
        
        subset = self.protein_id_type
        len_orig_set = len(self.orig_data)
        duplicates = self.orig_data[self.orig_data.duplicated(subset=subset)]
        self.orig_data.drop_duplicates(subset=subset, inplace=True)
        self.proteins = self.orig_data[self.protein_id_type]
        if not silent:
            print(f'{len(duplicates)} of {len_orig_set} genes were dropped from original set')
            if len(duplicates) < 1:
                return duplicates
            elif len(duplicates) < 20:
                print('Dropped rows from original set:')
                print(duplicates[self.protein_id_type])
            elif len(duplicates) < 80:
                print('Dropped genes from original set:\n', *list(duplicates.protein_id_type))
        return duplicates

    def get_category_terms(self, category:str, term_type:str='id')->set:
        """
        Parameters
        ----------
        category : str
            Name of category
        term_type : str
            'id' or 'description'.
            id - returns terms IDs of category (for example, GO terms)
            description - returns Description of IDs of category

        Returns
        -------
        set
            set of terms
        """
        d_term = {'id': 'term', 'description': 'description'} # dict associate term_type and colnames of enrichment table
        valid_category = self._get_valid_category()
        Check_Value(category, valid_category, 'category')
        Check_Value(term_type, {'description', 'id'}, 'term_type')
        return set(self.enrichment[d_term[term_type]][self.enrichment.category == category])

    def get_enrichment(self):
        """
        Function performs enrichment analysis. Results store in self.enrichment
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self.enrichment = stringdb.get_enrichment(self.genes_mapped.queryItem) #get enrichment
        self.enrichment['enrich_score'] = self.enrichment.fdr.apply(lambda x: round(-log(x, 2), 1)) #get enrichment score

    def get_genes_of_term(self, term:str)-> list:
        """
        Parameters
        ----------
        term : str
            target GO term from column 'term' in enrichment table

        Returns
        -------
        list
            list of genes associated with target term
        """
        try:
            return self.enrichment.inputGenes[self.enrichment.term == term].to_list()[0].rstrip().strip().split(',')
        except: print('Term not found')

    def get_genes_by_localization(self, compartments: list, set_operation: str, save=False):
        """
        Function for getting proteins localized in target compartments. You also can do common set operations
        under compartments genes
        Example: get_genes_by_localization([Nucleus, Cytosol], 'union') return proteins localized in Nucleus or Cytosol

        Parameters
        ----------
        compartments : list
            list of compartments. Will be attention:
                1) Capitalization of letters matters. Get available compartment names by calling "get_components_list()".
                2) Order of compartments matter if you want to get sets difference.
        set_operation : str
            operation between sets. This means that the operations will be applied sequentially to all
            sets from the compartments.
            For example:
                get_genes_by_localization(['Nucleus', 'Cytosol'], 'difference') return just nucleus proteins,
                get_genes_by_localization(['Cytosol', 'Nucleus'], 'union') return cytosol and nucleus proteins.
                get_genes_by_localization(['all', 'Nucleus'], 'difference') return all proteins except nucleus proteins.

        Returns
        -------
        set of proteins localized in target compartments
        """
        # check set_operation and compartments error
        Check_Value(set_operation, {'union', 'intersection', 'difference', 'symmetric_difference'}, 'set_operation')
        components_list = self.get_category_terms('Component', term_type='description').union({'all'})
        for c in compartments:
            Check_Value(c, components_list, 'Compartments',
                        message='There is no such compartment. To display a list of available compartments, '
                                'call <<show_category_terms("Components")>>. '
                                'If you want to get all genes, use tag "all" in compartments list')

        # define common set operations
        def union(a: set, b: set) -> set:
            return a.union(b)

        def intersection(a: set, b: set) -> set:
            return a.intersection(b)

        def difference(a: set, b: set) -> set:
            return a.difference(b)

        def symmetric_difference(a: set, b: set) -> set:
            return a.symmetric_difference(b)

        operations = {'union': union, 'intersection': intersection, 'difference': difference,
                      'symmetric_difference': symmetric_difference}

        # choose all strings in enrichment data connected with 'Component'
        component_data = self.enrichment[self.enrichment.category == 'Component']

        # create location genes set and apply set_operation for each compartment gene set
        if compartments[0] == 'all':
            loc_genes = set(self.proteins)
        else:
            loc_genes = set(component_data.inputGenes[component_data.description == compartments[0]]
                            .to_list()[0].rstrip().strip().split(','))
        for i in range(1, len(compartments)):
            if compartments[i] == 'all':
                compartment_genes = set(self.proteins)
            else:
                compartment_genes = set(component_data.inputGenes[component_data.description == compartments[i]]
                                        .to_list()[0].rstrip().strip().split(','))
            loc_genes = operations[set_operation](loc_genes, compartment_genes)
        print(f'{len(loc_genes)} genes were founded\n')

        if save: # save genes in txt format (1 gene on 1 string)
            filename = 'Genes_' + '_'.join(compartments)
            if len(filename) > 45:
                filename = 'Genes_' + compartments[0][:62] + '_and_' + str(len(compartments) - 1) + '_compartments'
            filename += '_' + set_operation + '.txt'
            with open(filename, 'w+') as f:
                for term in loc_genes:
                    f.write(term + '\n')
            print(f'File {filename} successfully saved in {os.path.abspath(os.getcwd())}\n')

        #return list(loc_genes)
        return pd.DataFrame({self.protein_id_type: list(loc_genes)})

    @titler('MAPPING GENES IN STRING')
    def get_mapped(self, species=9606):
        """
        Parameters
        ----------
        species : int, optional
            ID of organism. For example, Human species=9606. Default is 9606.

        Returns
        -------
        None
        """

        self.genes_mapped = stringdb.get_string_ids(self.proteins, species=species)
        self.nomapped_genes, self.overmapped_genes = self._find_nomapped_genes()
        print(
            f'{len(self.genes_mapped.queryItem.unique())} of {len(set(self.proteins.unique()))} unique genes were mapped\n')
        if len(self.nomapped_genes) < 80:
            print('List of nomapped genes:\n', list(self.nomapped_genes))
        if len(self.overmapped_genes) < 80:
            print('List of overmapped genes:\n', list(self.overmapped_genes))

    def get_terms_prioretizing(self, category, sorting_results=True, ascending=False):
        
        """
        Function for prioretizing GO-terms from choosen category
        
        Parameters
        ----------
        obj : EnrichmentAnalysis
            EnrichmentAnalysis object
        category : str
            name of category

        Returns
        -------
        **pd.DataFrame** enrichment table of prioretized Terms, stored in *self.prior_enrichment[category]*
        """
        #check validness of category
        valid_category = self._get_valid_category()
        Check_Value(category, valid_category, 'category')
        
        GO_terms = self.get_category_terms(category)
        prior_GO_terms = self.prioretizingGO(GO_terms)
        try:
            self.prior_enrichment[category] = create_subframe_by_names(self.enrichment, column='term', names=prior_GO_terms)
        except:
            self.prior_enrichment = {}
            self.prior_enrichment[category] = create_subframe_by_names(self.enrichment, column='term', names=prior_GO_terms)
        if sorting_results:
            self.prior_enrichment[category].sort_values(by=['enrich_score'], ascending=ascending, inplace=True)
        return self.prior_enrichment[category]
        

    def prioretizingGO(self, terms: [list, set], organism='Human', domain='BP'):
        """
        Function for prioretizing GO-terms using R script with GOxploreR package (doi:10.1038/s41598-020-73326-3)
        See 'RScript Prioretizing_GO.R'
        work with R.4-3.x. You need to add RScript in PATH

        If you use this function in google-collab, you will have to install R-packages at the first launch.
        This may take a long time (up to 20 minutes)

        Parameters
        ----------
        terms : list or set
            list of GO-terms
        organism : str
            name of target organism
        domain : str
            name of domain in GO-graph. Available inputs: 'BP' - Biological Process
                                                     'CC' - Cellular Component
                                                     "MF" - Molecular Functions

        Returns
        -------
        list of Prioretized GO terms
        """
        valid_organisms = {"Homo Sapiens", "Human", "Rattus Norvegicus", "Rat", "Mus Musculus", "Mouse",
                           "Danio Rerio", "Zebrafish", "Caenorhabditis Elegans", "Worm", "Arabidopsis Thaliana",
                           "Cress", "Saccharomyces Cerevisiae", "Yeast", "Schizosaccharomyces Pombe",
                           "Fission Yeast", "Drosophila Melanogaster", "Fruit Fly", "Escherichia Coli", "E.Coli"}
        Check_Value(organism, valid_organisms, 'organism')
        Check_Value(domain, {'BP', 'MF', 'CC'}, 'domain')

        installing = Check_R_packages(CRAN_packages=["GOxploreR", "data.table", "BiocManager", "utils", "ggplot2"],
                         BiocManager_packages=["GO.db", "annotate", "biomaRt"])

        save_table(pd.DataFrame(terms, columns=['Term']), 'input_priority_terms.csv', saveformat='csv', index=False)

        # Request to CMD to execute RScript
        command = 'Rscript'
        path2script = os.path.join(RSCRIPTS_PATH, 'Prioretizing_GO.R')
        path2file = os.path.abspath('input_priority_terms.csv')

        # Variable number of args in a list
        args = [path2file, 'Human', 'BP']
        # Build subprocess command
        cmd = [command, path2script] + args
        # check_output will run the command and store to result
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        output, error = p.communicate()

        # PRINT R CONSOLE OUTPUT (ERROR OR NOT)
        if p.returncode == 0:
            if installing:
                print('All R-packages were installed successfully')
            s_output = short_R_output(output.decode("utf8")) # if all is OK, then makes short output
            print(f'R OUTPUT:\n {s_output}')
        else:
            print(f'R ERROR:\n {error.decode("utf8")}')

        prior_terms = pd.read_csv(os.path.abspath('output_priority_terms.csv'))
        return list(prior_terms.Term)

    def proteins_participation_in_the_category(self, df, category, term_type='id', term_sep='\n'):
        """
        Function check terms that proteins participated and make statistics table

        Parameters
        ----------
        df : pd.DataFrame
            target DataFrame
        category : str
            Name of category
        term_type : str
            'id' or 'description'.
            id - returns terms IDs of category (for example, GO terms)
            description - returns Description of IDs of category
        term_sep : str
            terms connected with each protein will save in one cell. Choose separator beetwen terms

        Returns
        -------
        pd.DataFrame
            table with protein`s participation in category
        """
        d_term = {'id': 'term', 'description': 'description'} # dict associate term_type and colnames of enrichment table
        valid_category = self._get_valid_category()
        Check_Value(category, valid_category, 'category')
        Check_Value(term_type, {'description', 'id'}, 'term_type')

        prot_participation = pd.DataFrame(columns=[self.protein_id_type, 'number_of_terms', 'terms'])
        sub_df_category = df[df.category == category]
        for prot in self.proteins:
            prot_participation.loc[len(prot_participation)] = {
                self.protein_id_type: prot,
                'number_of_terms': len(sub_df_category[sub_df_category.inputGenes.str.contains(prot)]),
                'terms': sub_df_category[d_term[term_type]][sub_df_category.inputGenes.str.contains(prot)].apply(
                    lambda x: str(x) + term_sep).sum()}

        prot_participation.sort_values('number_of_terms', ascending=False, inplace=True)
        return prot_participation

    @titler('CATEGORY TERMS')
    def show_category_terms(self, category:str, show:[int, str]=10, sort_by='genes',
                            save:bool = False, savename='terms', saveformat='xlsx')->None:
        """
        Function displays all terms and number of associated genes in category

        Parameters
        ----------
        category : str
            Name of category. You can check available category by calling 'show_enrichment_categories' method
        show : int or str
            "all" or integer number. Number of strings to display
        sort_by : str
            ["genes", "term"] - sort by number of genes (by descending) or term names (by ascending)
        save : bool
            Need to save? Choose True. By default, save in .xlsx format
        savename : str
            work with save=True, name of file
        saveformat : str
            format of saving file: 'xlsx' or 'csv'
        """

        if type(show) != int and show != 'all':
            raise Exception('Error of "show" variable. Choose "all" or integer number')
        valid_category = self._get_valid_category()
        Check_Value(category, valid_category, 'category')
        Check_Value(sort_by, {'genes', 'term'}, 'sort_by')

        table = []
        category_data = self.enrichment[self.enrichment.category == category]
        terms = self.get_category_terms(category, term_type='description')
        
        for term in terms:
            string = category_data[category_data.description == term]
            table.append([term, list(string.number_of_genes)[0]])
        df = pd.DataFrame(table, columns=['Term', '# Genes'])
        if sort_by == 'genes':
            df.sort_values('# Genes', ascending = False, inplace=True)
        elif sort_by == 'term':
            df.sort_values('Term', ascending = True, inplace=True)
        if show == 'all':
            display_df(df)
        else:
            display_df(df.head(show))

        if save:
            if savename == 'terms':
                savename = category + '_' + savename + '_' + datetime.now().strftime('%m-%d-%Y')
            self.save_table(df, savename, saveformat=saveformat, index=False)

    @titler('ENRICHMENT CATEGORIES')
    def show_enrichment_categories(self):
        """
        function shown available enrichment categories for current dataset
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        table = []
        for term in self.enrichment.category.unique():
            table.append([term, len(self.enrichment[self.enrichment.category == term])])
        print(tabulate(table, headers=['Category', 'Number of terms'], tablefmt='orgtbl'))

    def show_enrichest_terms_in_category(self, category: str, count: [int, str]='all', sort_by='fdr',
                                         save: bool = False, savename='enrichment', saveformat='xlsx'):
        """
        Function shows top-%count of most enriched terms in %category

        Parameters
        ----------
        category : str
            Name of category. You can check available category by calling 'show_enrichment_categories' method
        count : int, str
            count of terms you need to show. Choose "all" or integer number
        sort_by : str
            you can sort target list by one of 'fdr', 'p_value', 'number_of_genes' parameters
        save : bool
            Need to save? Choose True. By default, save in .xlsx format
        savename : str
            work with save=True, name of file
        saveformat : str
            format of saving file: 'xlsx' or 'csv'
        Returns
        -------
        None
        """
        
        if type(count) != int and count != 'all':
            raise Exception('Error of "count" variable. Choose "all" or integer number')
        valid_category = self._get_valid_category()
        Check_Value(category, valid_category, 'category')
        Check_Value(sort_by, {'fdr', 'p_value', 'number_of_genes'}, 'sort_by')

        table = self.enrichment[self.enrichment.category == category].sort_values(by=sort_by)
        if count == 'all':
            count = len(table)
            
        if save:
            if savename == 'enrichment':
                savename += '_' + category + '_' + datetime.now().strftime('%m-%d-%Y')
            save_table(table.head(count), savename, saveformat=saveformat, index=False)
        print(f'ENRICHEST TERMS IN CATEGORY "{category}"')
        display_df(table.head(count).drop(['number_of_genes_in_background', 'ncbiTaxonId', 'preferredNames', 'p_value'], axis=1))
        return table


    # @staticmethod
    # def create_subframe_by_names(df, column: str, names: [list, tuple, set], add: str = 'first'):
    #     """
    #     Function finds rows in original dataset and returns sub-dataframe including input names in selected column

    #     Parameters
    #     ----------
    #     df : pd.DataFrame
    #         target DataFrame
    #     column : str
    #         the selected column in which names will be searched
    #     names : list of str
    #         list of target names whose records need to be found in the table
    #     add : str
    #         ['first', 'last', 'all'] parameter of adding found rows.
    #         'first' - add only the first entry
    #         'last' - add only the last entry
    #         'all' - add all entries

    #     Returns
    #     -------
    #     pd.DataFrame
    #         sub-dataframe including input names in selected column
    #     """
        
    #     Check_Value(add, {'first', 'last', 'all'}, add)

    #     def add_all(table, rows):
    #         return pd.concat([table, rows])

    #     def add_first(table, rows):
    #         table.loc[len(table)] = rows.iloc[0]
    #         return table

    #     def add_last(table, rows):
    #         table.loc[len(table)] = rows.iloc[-1]
    #         return table

    #     adding_method = {'first': add_first,
    #                      'last': add_last,
    #                      'all': add_all}

    #     new_df = pd.DataFrame(columns=df.columns)
    #     not_found_names = []
    #     for name in names:
    #         rows = df[df[column] == name]
    #         if len(rows) > 0:
    #             new_df = adding_method[add](new_df, rows)
    #         else: not_found_names.append(name)
    #     print(f'{len(not_found_names)} names were not found in the dataframe:\n')
    #     print('[', end='')
    #     print(*not_found_names, sep=', ', end='')
    #     print(']')

    #     return new_df


from   .wrappers import Check_Value, display_df, titler, print_upline, print_downline

from   datetime import datetime
import functools
from   math import log
import os
import pandas as pd
import stringdb
from   subprocess import Popen, PIPE
from   tabulate import tabulate


# path to directory contains all RScripts
RSCRIPTS_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'RScripts')


def get_mapping(ids, return_nomapped=False, protein_id_type='Gene', species=9606):
    """
    Mapping protein IDs to STRING IDs

    Parameters
    ----------
    ids : pd.DataFrame
        pd.DataFrame contains ids in *protein_id_type* column
    return_nomapped : bool
        return list of not mapped genes
    protein_id_type : str
        type of protein ID. Valid Types: 'UniProtID', 'Gene'
    species : int
        species id, for example, human species id = 9606

    Returns
    -------
    mapped_genes : list
        list of mapped genes (or list of not mapped genes if return_nomapped = True)
    """
    
    obj = STRING_mapping(ids, protein_id_type)
    obj.drop_duplicated_genes(silent=True)
    obj.get_mapped(species)
    if return_nomapped:
        return obj.genes_mapped, obj.nomapped_genes
    else:
        return obj.genes_mapped


class STRING_mapping():
    types = {'UniProtID': 'queryItem', 'Gene': 'preferredName'}

    def __init__(self, data, protein_id_type='Gene'):
        """
        Mapping class conctructor.
        
        Parameters
        ----------
        data : pd.DataFrame
            Dataframe containing the protein ID for analysis. It must contain either a "Gene" or "UniProtID" column
        protein_id_type : str
            type of protein ID. Valid Types

        Returns
        -------
        None
        """
        #check correctness of inputs
        self.protein_id_type = protein_id_type
        self._check_proteins_column(data)

        self.orig_data = data
        self.proteins = self.orig_data[self.protein_id_type]

    def _check_proteins_column(self, data):
        """
        The function checks the presence of the 'Gene' and 'UniProtID' columns in data

        Parameters
        ----------
        data : pd.DataFrame
            data contains proteins` IDs

        Returns
        -------
        None
        """

        valid_cols = set(self.types.keys()).intersection(data.columns)
        if len(valid_cols) == 0:
            message = 'The protein data must contain either a "Gene" or "UniProtID" column'
            del self
            raise Exception(message)
        elif len(valid_cols) == 1:
            p_id = next(iter(valid_cols))
            if self.protein_id_type != p_id:
                self.protein_id_type = p_id
                print(f'You choose "protein_id_type" that wasn`t contained in your data. '
                      f'"protein_id_type" is changed to "{self.protein_id_type}"\n')
        elif len(valid_cols) == 2:
            pass
        
    def _find_nomapped_genes(self):
        """
        Function for finding genes in dataset which didn't find by STRING (nomapped genes) and
        found by STRING but wasn't in dataset (overmapped genes)

        Parameters
        ----------
        None

        Returns
        -------
        tuple
            tuple of two lists. First list contains nomapped genes and second list contains overmapped genes.
        """

        nomapped = sorted(
            set(self.proteins.unique()).difference(set(self.genes_mapped[self.types[self.protein_id_type]].unique())))
        overmapped = sorted(
            set(self.genes_mapped[self.types[self.protein_id_type]].unique()).difference(set(self.proteins.unique())))
        return nomapped, overmapped

    @titler('DISCARDING DUPLICATES')
    def drop_duplicated_genes(self, silent=False):
        """
        Function for dropping duplicated genes.

        Parameters
        ----------
        silent : bool
            If True, function will not print information about dropped genes

        Returns
        -------
        DataFrame
            DataFrame with dropped genes
        """
        subset = self.protein_id_type
        len_orig_set = len(self.orig_data)
        duplicates = self.orig_data[self.orig_data.duplicated(subset=subset)]
        self.orig_data.drop_duplicates(subset=subset, inplace=True)
        self.proteins = self.orig_data[self.protein_id_type]
        if not silent:
            print(f'{len(duplicates)} of {len_orig_set} genes was dropped from original set')
            if len(duplicates) < 1:
                return duplicates
            elif len(duplicates) < 20:
                print('Dropped rows from original set:')
                print(duplicates[self.protein_id_type])
            elif len(duplicates) < 80:
                print('Dropped genes from original set:\n', *list(duplicates.protein_id_type))
        return duplicates
    
    @titler('MAPPING GENES IN STRING')
    def get_mapped(self, species=9606):
        """
        Function for mapping genes in STRING database

        Parameters
        ----------
        species : int
            ID of organism. For example, Human species=9606

        Returns
        -------
        None
        """

        self.genes_mapped = stringdb.get_string_ids(self.proteins, species=species)
        self.nomapped_genes, self.overmapped_genes = self._find_nomapped_genes()
        print(
            f'{len(self.genes_mapped.queryItem.unique())} of {len(set(self.proteins.unique()))} unique genes were mapped\n')
        if len(self.nomapped_genes) < 80:
            print('List of nomapped genes:\n', list(self.nomapped_genes))
        if len(self.overmapped_genes) < 80:
            print('List of overmapped genes:\n', list(self.overmapped_genes))
                
    @staticmethod
    def create_subframe_by_names(df, column: str, names: [list, tuple, set], add: str = 'first'):
        """
        Function finds rows in original dataset and returns sub-dataframe including input names in selected column

        Parameters
        ----------
        df : pd.DataFrame
            target DataFrame
        column : str
            the selected column in which names will be searched
        names : list of str
            list of target names whose records need to be found in the table
        add : str
            ['first', 'last', 'all'] parameter of adding found rows.
            'first' - add only the first entry
            'last' - add only the last entry
            'all' - add all entries

        Returns
        -------
        pd.DataFrame
            sub-dataframe including input names in selected column
        """
        
        Check_Value(add, {'first', 'last', 'all'}, add)

        def add_all(table, rows):
            return pd.concat([table, rows])

        def add_first(table, rows):
            table.loc[len(table)] = rows.iloc[0]
            return table

        def add_last(table, rows):
            table.loc[len(table)] = rows.iloc[-1]
            return table

        adding_method = {'first': add_first,
                         'last': add_last,
                         'all': add_all}

        new_df = pd.DataFrame(columns=df.columns)
        not_found_names = []
        for name in names:
            rows = df[df[column] == name]
            if len(rows) > 0:
                new_df = adding_method[add](new_df, rows)
            else: not_found_names.append(name)
        print(f'{len(not_found_names)} names were not found in the dataframe:\n')
        print('[', end='')
        print(*not_found_names, sep=', ', end='')
        print(']')

        return new_df



import functools
import pandas as pd
import os

def display_df(df):
    """
    Function for displaying DataFrames (df). If IPNB is used, df will display with common IPNB function 'Display', else:
    it will display by print() function

    Parameters
    ----------
    df : DataFrame
        DataFrame to display

    Returns
    -------
    None
    """
    try: display(df)
    except: print(df)

def Check_Value(val:[str, float, int], valid_values:set, valname:str, message='Wrong value123'):
    """
    Function for check correctness of input value

    Parameters
    ----------
    val : str, float, int
        input value
    valid_values : set
        set of valid values
    valname : str
        group name of valid_values set. Or name of val variable
    message : str
        Error message

    Returns
    -------
    None
    """
    if val not in valid_values:
        if message == 'Wrong value123':
            message = f'Wrong value of "{valname}" variable! Choose one of {valid_values}'
        raise Exception(message)

def Check_type(val, valid_types:tuple, valname:str, message='Wrong value123'):
    
    """
    Function for check correctness of input type

    Parameters
    ----------
    val : any
        input value
    valid_types : tuple
        tuple of valid types
    valname : str
        group name of valid_values set. Or name of val variable
    message : str
        optional. Error message

    Returns
    -------
    None
    """
    if not isinstance(val, valid_types):
        if message == 'Wrong value123':
            message = f'Wrong type of "{valname}" variable! Valid types are one of {valid_types}'
        raise Exception(message)
    
    
def Check_kwargs(kwargs, valid_params:set):
    """
    Function for check correctness of input kwargs
    Parameters
    ----------
    kwargs : dict
        dict of kwargs
    Returns
    -------
    None
    """
    for param in kwargs.keys():
        if param not in valid_params:
            raise ValueError(f' {param} parameter not valid. List of valid parameters: {valid_params}')

def print_downline(line_length:int=40):
    """
    Function prints line for titler decorator

    Parameters
    ----------
    line_length : int
        length of line (number of '_' symbols)

    Returns
    -------
        None
    """
    line = '_'*line_length
    print(f'{line}\n\n')

def print_upline(title:str, line_length:int=40):
    """
    function prints line and adds title for titler decorator

    Parameters
    ----------
    line_length : int
        length of line (number of '_' symbols)

    Returns
    -------
        None
    """
    line = '_'*line_length
    print(f'\t{title}\n{line}')

def titler(title: str, line_length=40):
    """
    Decorator added Title and edges of Paragraph

    Parameters
    ----------
    title : str
        title
    line_length : int
        length of line (number of '_' symbols)

    Returns
    -------
    callable
        decorated function
    """
    def titler_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            print_upline(title, line_length)
            original_result = func(*args, **kwargs)
            print_downline(line_length)

            return original_result
        return wrapper
    return titler_decorator

def save_table(table, name, saveformat='xlsx', index:bool = True):
    """
    Function for saving DataFrame tables
    
    Parameters
    ----------
    table : DataFrame
        DataFrame table to save
    name : str
        name of file
    saveformat : str
        format of saving file: 'xlsx' or 'csv'
    index : bool
        show indexes in saved table?
    
    Returns
    -------
    None
    """

    Check_Value(saveformat, {'csv', 'xlsx'}, 'saveformat')
    Check_type(table, pd.DataFrame, 'table')
    try:
        if saveformat == 'xlsx':
            if name[-5:] != '.xlsx' and name[-4:] != '.xls':
                name += '.xlsx'
            table.to_excel(name, index=index)
        elif saveformat == 'csv':
            if name[-4:] != '.csv':
                name += '.csv'
            table.to_csv(name, index=index, header=True)
        print(f'File {name} successfully saved in {os.path.abspath(os.getcwd())}\n')
    except PermissionError:
        print('Permission Denied Error: Access is denied. Close file if it`s open and try again')
    except:
        print('Saving file isn`t complete. If you rewrite file, close it and try again')
        
        
def create_subframe_by_names(df, column: str, names: [list, tuple, set], add: str = 'first'):
    """
    Function finds rows in original dataset and returns sub-dataframe including input names in selected column

    Parameters
    ----------
    df : pd.DataFrame
        target DataFrame
    column : str
        the selected column in which names will be searched
    names : list of str
        list of target names whose records need to be found in the table
    add : str
        ['first', 'last', 'all'] parameter of adding found rows.
        'first' - add only the first entry
        'last' - add only the last entry
        'all' - add all entries

    Returns
    -------
    pd.DataFrame
        sub-dataframe including input names in selected column
    """
    
    Check_Value(add, {'first', 'last', 'all'}, add)

    def add_all(table, rows):
        return pd.concat([table, rows])

    def add_first(table, rows):
        table.loc[len(table)] = rows.iloc[0]
        return table

    def add_last(table, rows):
        table.loc[len(table)] = rows.iloc[-1]
        return table

    adding_method = {'first': add_first,
                        'last': add_last,
                        'all': add_all}

    new_df = pd.DataFrame(columns=df.columns)
    not_found_names = []
    for name in names:
        rows = df[df[column] == name]
        if len(rows) > 0:
            new_df = adding_method[add](new_df, rows)
        else: not_found_names.append(name)
    print(f'{len(not_found_names)} names were not found in the dataframe:\n')
    print('[', end='')
    print(*not_found_names, sep=', ', end='')
    print(']')

    return new_df



import os
import pandas as pd
from   subprocess import Popen, PIPE


# path to directory contains all RScripts
RSCRIPTS_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'RScripts')

def Check_R_packages(CRAN_packages:None, BiocManager_packages:None) -> bool:
    """
    function checks target R-packages among installed R-packages
    function makes CMD request to RScript
    :param CRAN_packages: list of CRAN packages to check
    :param BiocManager_packages: list of BiocManager packages to check
    :return:
    """
    # if CRAN_packages is None, then CRAN_packages = empty list
    CRAN_packages = CRAN_packages or []
    BiocManager_packages = BiocManager_packages or []

    s = ''
    for p in CRAN_packages:
        s += str(p) + ','  # s - temp string contained all target CRAN-package names

    if len(s) > 0:
        CRAN = '1'
        with open('temp_CRAN_packages.txt', 'w+') as f1:
            f1.write(s.rstrip(','))
            f1.write('\n')
    else:
        CRAN = '0'

    s = ''
    for p in BiocManager_packages:
        s += str(p) + ',' # s - temp string contained all target BiocManager-package names

    if len(s) > 0:
        BiocM = '1'
        with open('temp_BiocManager_packages.txt', 'w+') as f2:
            f2.write(s.rstrip(','))
            f2.write('\n')
    else:
        BiocM = '0'

    command = 'Rscript'
    path2script = os.path.join(RSCRIPTS_PATH, 'check_packages.R')

    # Variable number of args in a list
    args = [CRAN, BiocM]
    # Build subprocess command
    cmd = [command, path2script] + args
    # check_output will run the command and store to result
    p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    output, error = p.communicate()

    # PRINT R CONSOLE OUTPUT (ERROR OR NOT)
    if p.returncode == 0 and len(output) > 2:
        print(f'R OUTPUT:\n {output.decode("utf8")}')
        # TEMPORARY SOLUTION
        # Check for not installed packages message:
        if output.decode("utf8").find('Packages not found') != -1:
            installing = True
        else: installing = False
    else:
        if len(error)>0: print(f'R ERROR:\n {error.decode("utf8")}')
        installing = False

    try: os.remove('temp_CRAN_packages.txt')
    except: pass
    try: os.remove('temp_BiocManager_packages.txt')
    except: pass

    return installing

def short_R_output(text):
    """
    function for shortening R output string. It takes string as parameter and return new string which
    contains only lines that starts with '[1]'
    This allows you to remove system messages, leaving only messages transmitted via 'print'
    
    :param text: text to be shortened
    :return: shortened text
    """
    df = pd.DataFrame(text.split('\n'), columns=['text'])
    new_text = df.text[df.text.apply(lambda x: True if x.find('[1]') >= 0 else False )].apply(lambda x: x.strip('[1] ').rstrip('\n')).apply(
                    lambda x: str(x) + '\n').sum()
    #isn`t working, for some reason 'contains' includs extra lines
    #new_text = df.text[df.text.str.contains('[1]')].apply(lambda x: x.strip('[1] ').rstrip('\n')).apply(
    #                lambda x: str(x) + '\n').sum()
    return new_text


from setuptools import setup, find_packages

def readme():
  with open('README.md', 'r', encoding="utf-8") as f:
    return f.read()

setup(
  name='ProteinNetworks',
  version='0.1.3',
  author='Mokin Yakov',
  author_email='mokinyakov@mail.ru',
  description='Module for working with protein networks (gene ontology, enrichment, protein-protein interactions, etc.)',
  long_description=readme(),
  long_description_content_type='text/markdown',
  url='https://github.com/skewer33/ProteinNetworks.git',
  packages=find_packages(),
  install_requires=[
    'pandas>=2.0.1',
    'stringdb==0.1.5',
    'tabulate==0.9.0'],
  classifiers=[
    'Programming Language :: Python :: 3.12',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent'
  ],
  keywords='proteins interactions PPI networks enrichment STRINGdb Bioilogical-Processes Molecular-Functions Cellular-Components Gene-Ontology',
  project_urls={
    'Documentation': 'https://github.com/skewer33/ProteinNetworks/blob/main/README.md'
  },
  python_requires='>=3.7',
  include_package_data=True
)
